# Copyright 2018 Yelp
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from mrjob.examples.mr_most_used_word import MRMostUsedWord

from tests.job import run_job
from tests.sandbox import BasicTestCase


class MRMostUsedWordTestCase(BasicTestCase):

    def test_empty(self):
        self.assertEqual(run_job(MRMostUsedWord()), {})

    def test_ignore_stop_words(self):
        RAW_INPUT = b"""
        A Car Joke:
        When is a car not a car?
        When it turns into a driveway!
        """

        EXPECTED_OUTPUT = {
            None: u'car',
        }

        self.assertEqual(run_job(MRMostUsedWord(), RAW_INPUT),
                         EXPECTED_OUTPUT)



# -*- coding: utf-8 -*-
# Copyright 2018 Yelp
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for AMI utilities"""
from mrjob.ami import describe_base_emr_images

from tests.mock_boto3 import MockBoto3TestCase


class DescribeBaseEMRImagesTestCase(MockBoto3TestCase):

    # a valid base EMR image. we can make variants of this for testing
    BASE_EMR_IMAGE = {
        'Architecture': 'x86_64',
        'BlockDeviceMappings': [
            {
                'DeviceName': '/dev/xvda',
                'Ebs': {
                    'DeleteOnTermination': True,
                    'Encrypted': False,
                    'SnapshotId': 'snap-0ceb5dfba7c0cbd4c',
                    'VolumeSize': 8,
                    'VolumeType': 'standard'
                }
            }
        ],
        'CreationDate': '2018-08-11T02:33:53.000Z',
        'Description': 'Amazon Linux AMI 2018.03.0.20180811 x86_64 HVM EBS',
        'EnaSupport': True,
        'Hypervisor': 'xen',
        'ImageId': 'ami-09c6e771',
        'ImageLocation': 'amazon/amzn-ami-hvm-2018.03.0.20180811-x86_64-ebs',
        'ImageOwnerAlias': 'amazon',
        'ImageType': 'machine',
        'Name': 'amzn-ami-hvm-2018.03.0.20180811-x86_64-ebs',
        'OwnerId': '137112412989',
        'Public': True,
        'RootDeviceName': '/dev/xvda',
        'RootDeviceType': 'ebs',
        'SriovNetSupport': 'simple',
        'State': 'available',
        'VirtualizationType': 'hvm',
    }

    def make_image(self, **kwargs):
        """Return a copy of BASE_EMR_IMAGE with the given fields added.

        You can blank out fields by setting them to None."""
        image = dict(self.BASE_EMR_IMAGE, **kwargs)

        return {k: v for k, v in image.items() if v is not None}

    def test_no_images(self):
        self.assertEqual(describe_base_emr_images(self.client('ec2')), [])

    def test_base_emr_image(self):
        image = self.make_image()
        self.add_mock_ec2_image(image)

        self.assertEqual(describe_base_emr_images(self.client('ec2')),
                         [image])

    def test_most_recent_image_first(self):
        image_old = self.make_image(ImageId='ami-old',
                                    CreationDate='2010-06-06T00:00:00.000Z')
        image_new = self.make_image(ImageId='ami-new',
                                    CreationDate='2015-05-06T00:00:00.000Z')

        self.add_mock_ec2_image(image_old)
        self.add_mock_ec2_image(image_new)

        self.assertEqual(describe_base_emr_images(self.client('ec2')),
                         [image_new, image_old])

    def test_filter_and_sort(self):
        image_old = self.make_image(ImageId='ami-old',
                                    CreationDate='2010-06-06T00:00:00.000Z')
        image_new = self.make_image(ImageId='ami-new',
                                    CreationDate='2015-05-06T00:00:00.000Z')
        image_null = {}

        self.add_mock_ec2_image(image_null)
        self.add_mock_ec2_image(image_old)
        self.add_mock_ec2_image(image_null)
        self.add_mock_ec2_image(image_new)
        self.add_mock_ec2_image(image_null)

        self.assertEqual(describe_base_emr_images(self.client('ec2')),
                         [image_new, image_old])

    def assert_rejects_image(self, **kwargs):
        image = self.make_image(**kwargs)
        self.add_mock_ec2_image(image)

        self.assertNotIn(image, describe_base_emr_images(self.client('ec2')))

    def test_owner_must_be_amazon(self):
        self.assert_rejects_image(ImageOwnerAlias='aws-marketplace',
                                  OwnerId='679593333241')

    def test_architecture_must_be_x86_64(self):
        self.assert_rejects_image(Architecture='i386')

    def test_root_device_type_must_be_ebs(self):
        self.assert_rejects_image(RootDeviceType='instance-store')

    def test_virtualization_type_must_be_hvm(self):
        self.assert_rejects_image(VirtualizationType='paravirtual')

    def test_amazon_linux_1_only(self):
        self.assert_rejects_image(
            Name='amzn2-ami-hvm-2017.12.0.20180109-x86_64-ebs')

    def test_stable_amazon_linux_versions_only(self):
        # no "testing" or "rc," only dots and numbers, please
        self.assert_rejects_image(
            Name='amzn-ami-hvm-2017.03.rc-1.20170327-x86_64-ebs')

    def test_one_volume_only(self):
        self.assert_rejects_image(
            BlockDeviceMappings=[
                self.BASE_EMR_IMAGE['BlockDeviceMappings'][0],
                {
                    'DeviceName': 'xvdca',
                    'VirtualName': 'ephemeral0',
                }
            ]
        )

    def test_dont_crash_on_missing_name(self):
        # shouldn't happen in practice, but just in case
        self.assert_rejects_image(Name=None)

    def test_dont_crash_on_missing_block_device_mappings(self):
        # shouldn't happen in practice, but just in case
        self.assert_rejects_image(BlockDeviceMappings=None)

    def test_dont_crash_on_missing_creation_date(self):
        self.assert_rejects_image(CreationDate=None)



# Copyright 2011 Yelp
# Copyright 2012 Yelp and Contributors
# Copyright 2015-2017 Yelp
# Copyright 2018 Yelp
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Very basic tests for the audit_usage script"""
import sys
from datetime import date
from datetime import datetime
from datetime import timedelta

import boto3
from dateutil.parser import parse
from dateutil.tz import tzutc

from mrjob.tools.emr.audit_usage import _cluster_to_full_summary
from mrjob.tools.emr.audit_usage import _percent
from mrjob.tools.emr.audit_usage import _subdivide_interval_by_date
from mrjob.tools.emr.audit_usage import _subdivide_interval_by_hour
from mrjob.tools.emr.audit_usage import main

from tests.py2 import patch
from tests.sandbox import BasicTestCase
from tests.tools.emr import ToolTestCase


# this test used to use naive datetimes
def utc(*args):
    return datetime(*args, tzinfo=tzutc())


class AuditUsageTestCase(ToolTestCase):

    def setUp(self):
        super(AuditUsageTestCase, self).setUp()

        self.repeat_sleep = self.start(patch('time.sleep'))
        # this is called once per cluster (no pagination), so we can
        # test quantity as well as whether it was called
        self.describe_cluster_sleep = self.start(
            patch('mrjob.tools.emr.audit_usage.sleep'))

    def test_with_no_clusters(self):
        self.monkey_patch_stdout()
        main(['-q', '--no-conf'])  # make sure it doesn't crash

        self.assertTrue(self.repeat_sleep.called)
        self.assertFalse(self.describe_cluster_sleep.called)

    def test_with_one_cluster(self):
        emr_client = boto3.client('emr')
        emr_client.run_job_flow(
            Name='no name',
            Instances=dict(
                MasterInstanceType='m1.medium',
                InstanceCount=1),
            JobFlowRole='fake-instance-profile',
            ServiceRole='fake-service-role',
            ReleaseLabel='emr-5.0.0')

        self.monkey_patch_stdout()
        main(['-q', '--no-conf'])
        self.assertIn(b'j-MOCKCLUSTER0', sys.stdout.getvalue())

        self.assertTrue(self.repeat_sleep.called)
        self.assertEqual(self.describe_cluster_sleep.call_count, 1)


class ClusterToFullSummaryTestCase(BasicTestCase):

    maxDiff = None  # show whole diff when tests fail

    def test_basic_cluster_with_no_steps(self):
        cluster = dict(
            BootstrapActions=[],
            Id='j-ISFORJAGUAR',
            Name='mr_exciting.woo.20100605.235850.000000',
            NormalizedInstanceHours=10,
            Steps=[],
            Status=dict(
                State='TERMINATED',
                Timeline=dict(
                    CreationDateTime=parse('2010-06-06T00:00:00Z'),
                    EndDateTime=parse('2010-06-06T00:30:00Z'),
                    ReadyDateTime=parse('2010-06-06T00:15:00Z'),
                ),
            ),
        )

        summary = _cluster_to_full_summary(cluster)

        self.assertEqual(summary, {
            'created': utc(2010, 6, 6, 0, 0),
            'end': utc(2010, 6, 6, 0, 30),
            'id': u'j-ISFORJAGUAR',
            'label': u'mr_exciting',
            'name': u'mr_exciting.woo.20100605.235850.000000',
            'nih': 10.0,
            'nih_bbnu': 2.5,
            'nih_billed': 5.0,
            'nih_used': 2.5,  # everything after bootstrapping was wasted
            'num_steps': 0,
            'owner': u'woo',
            'pool': None,
            'ran': timedelta(minutes=30),
            'ready': utc(2010, 6, 6, 0, 15),
            'state': u'TERMINATED',
            'usage': [{
                'date_to_nih_bbnu': {date(2010, 6, 6): 2.5},
                'date_to_nih_billed': {date(2010, 6, 6): 5.0},
                'date_to_nih_used': {date(2010, 6, 6): 2.5},
                'end': utc(2010, 6, 6, 0, 15),
                'end_billing': utc(2010, 6, 6, 0, 30),
                'label': u'mr_exciting',
                'hour_to_nih_bbnu': {utc(2010, 6, 6, 0): 2.5},
                'hour_to_nih_billed': {utc(2010, 6, 6, 0): 5.0},
                'hour_to_nih_used': {utc(2010, 6, 6, 0): 2.5},
                'nih_bbnu': 2.5,
                'nih_billed': 5.0,
                'nih_used': 2.5,
                'owner': u'woo',
                'start': utc(2010, 6, 6, 0, 0),
                'step_num': None,
            }],
        })

    def test_still_running_cluster_with_no_steps(self):

        cluster = dict(
            BootstrapActions=[],
            Id='j-ISFORJUICE',
            Name='mr_exciting.woo.20100605.235850.000000',
            NormalizedInstanceHours=10,
            Steps=[],
            Status=dict(
                State='WAITING',
                Timeline=dict(
                    CreationDateTime=parse('2010-06-06T00:00:00Z'),
                    ReadyDateTime=parse('2010-06-06T00:15:00Z'),
                ),
            ),
        )

        summary = _cluster_to_full_summary(
            cluster, now=utc(2010, 6, 6, 0, 30))

        self.assertEqual(summary, {
            'created': utc(2010, 6, 6, 0, 0),
            'end': None,
            'id': u'j-ISFORJUICE',
            'label': u'mr_exciting',
            'name': u'mr_exciting.woo.20100605.235850.000000',
            'nih': 10.0,
            'nih_bbnu': 2.5,
            'nih_billed': 5.0,
            'nih_used': 2.5,
            'num_steps': 0,
            'owner': u'woo',
            'pool': None,
            'ran': timedelta(minutes=30),
            'ready': utc(2010, 6, 6, 0, 15),
            'state': u'WAITING',
            'usage': [{
                'date_to_nih_bbnu': {date(2010, 6, 6): 2.5},
                'date_to_nih_billed': {date(2010, 6, 6): 5.0},
                'date_to_nih_used': {date(2010, 6, 6): 2.5},
                'end': utc(2010, 6, 6, 0, 15),
                'end_billing': utc(2010, 6, 6, 0, 30),
                'hour_to_nih_bbnu': {utc(2010, 6, 6, 0): 2.5},
                'hour_to_nih_billed': {utc(2010, 6, 6, 0): 5.0},
                'hour_to_nih_used': {utc(2010, 6, 6, 0): 2.5},
                'label': u'mr_exciting',
                'nih_bbnu': 2.5,
                'nih_billed': 5.0,
                'nih_used': 2.5,
                'owner': u'woo',
                'start': utc(2010, 6, 6, 0, 0),
                'step_num': None,
            }],
        })

    def test_still_bootstrapping_cluster_with_no_steps(self):
        cluster = dict(
            BootstrapActions=[],
            Id='j-ISFORJOKE',
            Name='mr_exciting.woo.20100605.235850.000000',
            NormalizedInstanceHours=10,
            Status=dict(
                State='BOOTSTRAPPING',
                Timeline=dict(
                    CreationDateTime=parse('2010-06-06T00:00:00Z'),
                ),
            ),
            Steps=[],
        )

        summary = _cluster_to_full_summary(
            cluster, now=utc(2010, 6, 6, 0, 30))

        self.assertEqual(summary, {
            'created': utc(2010, 6, 6, 0, 0),
            'end': None,
            'id': u'j-ISFORJOKE',
            'label': u'mr_exciting',
            'name': u'mr_exciting.woo.20100605.235850.000000',
            'nih': 10.0,
            'nih_bbnu': 0.0,
            'nih_billed': 5.0,
            'nih_used': 5.0,
            'num_steps': 0,
            'owner': u'woo',
            'pool': None,
            'ran': timedelta(minutes=30),
            'ready': None,
            'state': u'BOOTSTRAPPING',
            'usage': [{
                'date_to_nih_bbnu': {},
                'date_to_nih_billed': {date(2010, 6, 6): 5.0},
                'date_to_nih_used': {date(2010, 6, 6): 5.0},
                'end': utc(2010, 6, 6, 0, 30),
                'end_billing': utc(2010, 6, 6, 0, 30),
                'hour_to_nih_bbnu': {},
                'hour_to_nih_billed': {utc(2010, 6, 6, 0): 5.0},
                'hour_to_nih_used': {utc(2010, 6, 6, 0): 5.0},
                'label': u'mr_exciting',
                'nih_bbnu': 0.0,
                'nih_billed': 5.0,
                'nih_used': 5.0,
                'owner': u'woo',
                'start': utc(2010, 6, 6, 0, 0),
                'step_num': None,
            }],
        })

    def test_cluster_that_was_terminated_before_ready(self):
        cluster = dict(
            BootstrapActions=[],
            Id='j-ISFORJOURNEY',
            Name='mr_exciting.woo.20100605.235850.000000',
            NormalizedInstanceHours=1,
            Steps=[],
            Status=dict(
                State='TERMINATED',
                Timeline=dict(
                    CreationDateTime=parse('2010-06-06T00:00:00Z'),
                    EndDateTime=parse('2010-06-06T00:30:00Z'),
                ),
            ),
        )

        summary = _cluster_to_full_summary(
            cluster, now=utc(2010, 6, 6, 1))

        self.assertEqual(summary, {
            'created': utc(2010, 6, 6, 0, 0),
            'end': utc(2010, 6, 6, 0, 30),
            'id': u'j-ISFORJOURNEY',
            'label': u'mr_exciting',
            'name': u'mr_exciting.woo.20100605.235850.000000',
            'nih': 1.0,
            'nih_bbnu': 0.0,
            'nih_billed': 0.5,
            'nih_used': 0.5,
            'num_steps': 0,
            'owner': u'woo',
            'pool': None,
            'ran': timedelta(minutes=30),
            'ready': None,
            'state': u'TERMINATED',
            'usage': [{
                'date_to_nih_bbnu': {},
                'date_to_nih_billed': {date(2010, 6, 6): 0.5},
                'date_to_nih_used': {date(2010, 6, 6): 0.5},
                'end': utc(2010, 6, 6, 0, 30),
                'end_billing': utc(2010, 6, 6, 0, 30),
                'hour_to_nih_bbnu': {},
                'hour_to_nih_billed': {utc(2010, 6, 6, 0, 0): 0.5},
                'hour_to_nih_used': {utc(2010, 6, 6, 0, 0): 0.5},
                'label': u'mr_exciting',
                'nih_bbnu': 0.0,
                'nih_billed': 0.5,
                'nih_used': 0.5,
                'owner': u'woo',
                'start': utc(2010, 6, 6, 0, 0),
                'step_num': None,
            }],
        })

    def test_cluster_with_no_steps_split_over_midnight(self):
        cluster = dict(
            BootstrapActions=[],
            Id='j-ISFORJOY',
            Name='mr_exciting.woo.20100605.232950.000000',
            NormalizedInstanceHours=20,
            Status=dict(
                State='TERMINATED',
                Timeline=dict(
                    CreationDateTime=parse('2010-06-05T23:30:00Z'),
                    EndDateTime=parse('2010-06-06T01:15:00Z'),
                    ReadyDateTime=parse('2010-06-05T23:45:00Z'),
                ),
            ),
            Steps=[],
        )

        summary = _cluster_to_full_summary(cluster)

        self.assertEqual(summary, {
            'created': utc(2010, 6, 5, 23, 30),
            'end': utc(2010, 6, 6, 1, 15),
            'id': u'j-ISFORJOY',
            'label': u'mr_exciting',
            'name': u'mr_exciting.woo.20100605.232950.000000',
            'nih': 20.0,
            'nih_bbnu': 15.0,
            'nih_billed': 17.5,
            'nih_used': 2.5,
            'num_steps': 0,
            'owner': u'woo',
            'pool': None,
            'ran': timedelta(hours=1, minutes=45),
            'ready': utc(2010, 6, 5, 23, 45),
            'state': u'TERMINATED',
            'usage': [{
                'date_to_nih_bbnu': {date(2010, 6, 5): 2.5,
                                     date(2010, 6, 6): 12.5},
                'date_to_nih_billed': {date(2010, 6, 5): 5.0,
                                       date(2010, 6, 6): 12.5},
                'date_to_nih_used': {date(2010, 6, 5): 2.5},
                'end': utc(2010, 6, 5, 23, 45),
                'end_billing': utc(2010, 6, 6, 1, 15),
                'hour_to_nih_bbnu': {utc(2010, 6, 5, 23): 2.5,
                                     utc(2010, 6, 6, 0): 10.0,
                                     utc(2010, 6, 6, 1): 2.5},
                'hour_to_nih_billed': {utc(2010, 6, 5, 23): 5.0,
                                       utc(2010, 6, 6, 0): 10.0,
                                       utc(2010, 6, 6, 1): 2.5},
                'hour_to_nih_used': {utc(2010, 6, 5, 23): 2.5},
                'label': u'mr_exciting',
                'nih_bbnu': 15.0,
                'nih_billed': 17.5,
                'nih_used': 2.5,
                'owner': u'woo',
                'start': utc(2010, 6, 5, 23, 30),
                'step_num': None,
            }],
        })

    def test_cluster_with_one_still_running_step(self):
        cluster = dict(
            BootstrapActions=[],
            Id='j-ISFORJUNGLE',
            Name='mr_exciting.woo.20100606.035855.000000',
            NormalizedInstanceHours=20,
            Status=dict(
                State='RUNNING',
                Timeline=dict(
                    CreationDateTime=parse('2010-06-06T04:00:00Z'),
                    ReadyDateTime=parse('2010-06-06T04:15:00Z'),
                ),
            ),
            Steps=[
                dict(
                    Name='mr_exciting.woo.20100606.035855.000000: Step 1 of 3',
                    Status=dict(
                        Timeline=dict(
                            StartDateTime=parse('2010-06-06T04:15:00Z'),
                        ),
                    ),
                ),
            ],
        )

        summary = _cluster_to_full_summary(
            cluster, now=utc(2010, 6, 6, 5, 30))

        self.assertEqual(summary, {
            'created': utc(2010, 6, 6, 4, 0),
            'end': None,
            'id': u'j-ISFORJUNGLE',
            'label': u'mr_exciting',
            'name': u'mr_exciting.woo.20100606.035855.000000',
            'nih': 20.0,
            'nih_bbnu': 0.0,
            'nih_billed': 15.0,
            'nih_used': 15.0,
            'num_steps': 1,
            'owner': u'woo',
            'pool': None,
            'ran': timedelta(hours=1, minutes=30),
            'ready': utc(2010, 6, 6, 4, 15),
            'state': u'RUNNING',
            'usage': [{
                # bootstrapping
                'date_to_nih_used': {date(2010, 6, 6): 2.5},
                'date_to_nih_bbnu': {},
                'date_to_nih_billed': {date(2010, 6, 6): 2.5},
                'end': utc(2010, 6, 6, 4, 15),
                'end_billing': utc(2010, 6, 6, 4, 15),
                'hour_to_nih_used': {utc(2010, 6, 6, 4): 2.5},
                'hour_to_nih_bbnu': {},
                'hour_to_nih_billed': {utc(2010, 6, 6, 4): 2.5},
                'label': u'mr_exciting',
                'nih_used': 2.5,
                'nih_bbnu': 0.0,
                'nih_billed': 2.5,
                'owner': u'woo',
                'start': utc(2010, 6, 6, 4, 0),
                'step_num': None,
            }, {
                # mr_exciting, step 1
                'date_to_nih_used': {date(2010, 6, 6): 12.5},
                'date_to_nih_bbnu': {},
                'date_to_nih_billed': {date(2010, 6, 6): 12.5},
                'end': utc(2010, 6, 6, 5, 30),
                'end_billing': utc(2010, 6, 6, 5, 30),
                'hour_to_nih_used': {utc(2010, 6, 6, 4): 7.5,
                                     utc(2010, 6, 6, 5): 5.0},
                'hour_to_nih_bbnu': {},
                'hour_to_nih_billed': {utc(2010, 6, 6, 4): 7.5,
                                       utc(2010, 6, 6, 5): 5.0},
                'label': u'mr_exciting',
                'nih_used': 12.5,
                'nih_bbnu': 0.0,
                'nih_billed': 12.5,
                'owner': u'woo',
                'start': utc(2010, 6, 6, 4, 15),
                'step_num': 1,
            }],
        })

    def test_cluster_with_one_cancelled_step(self):
        cluster = dict(
            BootstrapActions=[],
            Id='j-ISFORJACUZZI',
            Name='mr_exciting.woo.20100606.035855.000000',
            NormalizedInstanceHours=20,
            Status=dict(
                State='RUNNING',
                Timeline=dict(
                    CreationDateTime=parse('2010-06-06T04:00:00Z'),
                    EndDateTime=parse('2010-06-06T05:30:00Z'),
                    ReadyDateTime=parse('2010-06-06T04:15:00Z'),
                ),
            ),
            # step doesn't have end time even though cluster does
            Steps=[
                dict(
                    Name='mr_exciting.woo.20100606.035855.000000: Step 1 of 3',
                    Status=dict(
                        Timeline=dict(
                            StartDateTime=parse('2010-06-06T04:15:00Z'),
                        ),
                    ),
                ),
            ]
        )

        summary = _cluster_to_full_summary(cluster)

        self.assertEqual(summary, {
            'created': utc(2010, 6, 6, 4, 0),
            'end': utc(2010, 6, 6, 5, 30),
            'id': u'j-ISFORJACUZZI',
            'label': u'mr_exciting',
            'name': u'mr_exciting.woo.20100606.035855.000000',
            'nih': 20.0,
            'nih_bbnu': 12.5,
            'nih_billed': 15.0,
            'nih_used': 2.5,
            'num_steps': 1,
            'owner': u'woo',
            'pool': None,
            'ran': timedelta(hours=1, minutes=30),
            'ready': utc(2010, 6, 6, 4, 15),
            'state': u'RUNNING',
            'usage': [{
                # bootstrapping
                'date_to_nih_used': {date(2010, 6, 6): 2.5},
                'date_to_nih_bbnu': {},
                'date_to_nih_billed': {date(2010, 6, 6): 2.5},
                'end': utc(2010, 6, 6, 4, 15),
                'end_billing': utc(2010, 6, 6, 4, 15),
                'hour_to_nih_used': {utc(2010, 6, 6, 4): 2.5},
                'hour_to_nih_bbnu': {},
                'hour_to_nih_billed': {utc(2010, 6, 6, 4): 2.5},
                'label': u'mr_exciting',
                'nih_used': 2.5,
                'nih_bbnu': 0.0,
                'nih_billed': 2.5,
                'owner': u'woo',
                'start': utc(2010, 6, 6, 4, 0),
                'step_num': None,
            }, {
                # mr_exciting, step 1 (cancelled)
                'date_to_nih_used': {},
                'date_to_nih_bbnu': {date(2010, 6, 6): 12.5},
                'date_to_nih_billed': {date(2010, 6, 6): 12.5},
                'end': utc(2010, 6, 6, 4, 15),
                'end_billing': utc(2010, 6, 6, 5, 30),
                'hour_to_nih_used': {},
                'hour_to_nih_bbnu': {utc(2010, 6, 6, 4): 7.5,
                                     utc(2010, 6, 6, 5): 5.0},
                'hour_to_nih_billed': {utc(2010, 6, 6, 4): 7.5,
                                       utc(2010, 6, 6, 5): 5.0},
                'label': u'mr_exciting',
                'nih_used': 0.0,
                'nih_bbnu': 12.5,
                'nih_billed': 12.5,
                'owner': u'woo',
                'start': utc(2010, 6, 6, 4, 15),
                'step_num': 1,
            }],
        })

    def test_multi_step_cluster(self):
        cluster = dict(
            BootstrapActions=[],
            Id='j-ISFORJOB',
            Name='mr_exciting.woo.20100605.232850.000000',
            NormalizedInstanceHours=20,
            Status=dict(
                State='TERMINATED',
                Timeline=dict(
                    CreationDateTime=parse('2010-06-05T23:30:00Z'),
                    EndDateTime=parse('2010-06-06T01:15:00Z'),
                    ReadyDateTime=parse('2010-06-05T23:45:00Z'),
                ),
            ),
            Steps=[
                dict(
                    Name='mr_exciting.woo.20100605.232850.000000: Step 1 of 3',
                    Status=dict(
                        Timeline=dict(
                            StartDateTime=parse('2010-06-05T23:45:00Z'),
                            EndDateTime=parse('2010-06-06T00:15:00Z'),
                        ),
                    ),
                ),
                dict(
                    Name='mr_exciting.woo.20100605.232850.000000: Step 2 of 3',
                    Status=dict(
                        Timeline=dict(
                            StartDateTime=parse('2010-06-06T00:30:00Z'),
                            EndDateTime=parse('2010-06-06T00:45:00Z'),
                        ),
                    ),
                ),
                dict(
                    Name='mr_exciting.woo.20100605.232850.000000: Step 3 of 3',
                    Status=dict(
                        Timeline=dict(
                            StartDateTime=parse('2010-06-06T00:45:00Z'),
                            EndDateTime=parse('2010-06-06T01:00:00Z'),
                        ),
                    ),
                ),
            ],
        )

        summary = _cluster_to_full_summary(cluster)

        self.assertEqual(summary, {
            'created': utc(2010, 6, 5, 23, 30),
            'end': utc(2010, 6, 6, 1, 15),
            'id': u'j-ISFORJOB',
            'label': u'mr_exciting',
            'name': u'mr_exciting.woo.20100605.232850.000000',
            'nih': 20.0,
            'nih_bbnu': 5.0,
            'nih_billed': 17.5,
            'nih_used': 12.5,
            'num_steps': 3,
            'owner': u'woo',
            'pool': None,
            'ran': timedelta(hours=1, minutes=45),
            'ready': utc(2010, 6, 5, 23, 45),
            'state': u'TERMINATED',
            'usage': [{
                # bootstrapping
                'date_to_nih_used': {date(2010, 6, 5): 2.5},
                'date_to_nih_bbnu': {},
                'date_to_nih_billed': {date(2010, 6, 5): 2.5},
                'end': utc(2010, 6, 5, 23, 45),
                'end_billing': utc(2010, 6, 5, 23, 45),
                'hour_to_nih_bbnu': {},
                'hour_to_nih_billed': {utc(2010, 6, 5, 23): 2.5},
                'hour_to_nih_used': {utc(2010, 6, 5, 23): 2.5},
                'label': u'mr_exciting',
                'nih_used': 2.5,
                'nih_bbnu': 0.0,
                'nih_billed': 2.5,
                'owner': u'woo',
                'start': utc(2010, 6, 5, 23, 30),
                'step_num': None,
            }, {
                # step 1 (and idle time after)
                'date_to_nih_used': {date(2010, 6, 5): 2.5,
                                     date(2010, 6, 6): 2.5},
                'date_to_nih_bbnu': {date(2010, 6, 6): 2.5},
                'date_to_nih_billed': {date(2010, 6, 5): 2.5,
                                       date(2010, 6, 6): 5.0},
                'end': utc(2010, 6, 6, 0, 15),
                'end_billing': utc(2010, 6, 6, 0, 30),
                'hour_to_nih_bbnu': {utc(2010, 6, 6, 0): 2.5},
                'hour_to_nih_billed': {utc(2010, 6, 5, 23): 2.5,
                                       utc(2010, 6, 6, 0): 5.0},
                'hour_to_nih_used': {utc(2010, 6, 5, 23): 2.5,
                                     utc(2010, 6, 6, 0): 2.5},
                'label': u'mr_exciting',
                'nih_used': 5.0,
                'nih_bbnu': 2.5,
                'nih_billed': 7.5,
                'owner': u'woo',
                'start': utc(2010, 6, 5, 23, 45),
                'step_num': 1,
            }, {
                # step 2
                'date_to_nih_used': {date(2010, 6, 6): 2.5},
                'date_to_nih_bbnu': {},
                'date_to_nih_billed': {date(2010, 6, 6): 2.5},
                'end': utc(2010, 6, 6, 0, 45),
                'end_billing': utc(2010, 6, 6, 0, 45),
                'hour_to_nih_bbnu': {},
                'hour_to_nih_billed': {utc(2010, 6, 6, 0): 2.5},
                'hour_to_nih_used': {utc(2010, 6, 6, 0): 2.5},
                'label': u'mr_exciting',
                'nih_used': 2.5,
                'nih_bbnu': 0.0,
                'nih_billed': 2.5,
                'owner': u'woo',
                'start': utc(2010, 6, 6, 0, 30),
                'step_num': 2,
            }, {
                # step 3 (and idle time after)
                'date_to_nih_used': {date(2010, 6, 6): 2.5},
                'date_to_nih_bbnu': {date(2010, 6, 6): 2.5},
                'date_to_nih_billed': {date(2010, 6, 6): 5.0},
                'end': utc(2010, 6, 6, 1, 0),
                'end_billing': utc(2010, 6, 6, 1, 15),
                'hour_to_nih_bbnu': {utc(2010, 6, 6, 1): 2.5},
                'hour_to_nih_billed': {utc(2010, 6, 6, 0): 2.5,
                                       utc(2010, 6, 6, 1): 2.5},
                'hour_to_nih_used': {utc(2010, 6, 6, 0): 2.5},
                'label': u'mr_exciting',
                'nih_used': 2.5,
                'nih_bbnu': 2.5,
                'nih_billed': 5.0,
                'owner': u'woo',
                'start': utc(2010, 6, 6, 0, 45),
                'step_num': 3,
            }],
        })

    def test_pooled_cluster(self):
        self._test_new_or_legacy_pooled_cluster(Tags=[
            dict(Key='__mrjob_pool_hash',
                 Value='0123456789abcdef0123456789abcdef'),
            dict(Key='__mrjob_pool_name',
                 Value='reflecting'),
        ])

    def test_legacy_pooled_cluster(self):
        # audit clusters from previous versions of mrjob
        self._test_new_or_legacy_pooled_cluster(
            BootstrapActions=[
                dict(Args=[], Name='empty'),
                dict(Args=['pool-0123456789abcdef0123456789abcdef',
                           'reflecting'], Name='master'),
            ],
        )

    def _test_new_or_legacy_pooled_cluster(self, **kwargs):
        # same as test case above with different job keys
        cluster = dict(
            Id='j-ISFORJOB',
            Name='mr_exciting.woo.20100605.232850.000000',
            NormalizedInstanceHours=20,
            State='TERMINATED',
            Status=dict(
                State='TERMINATED',
                Timeline=dict(
                    CreationDateTime=parse('2010-06-05T23:30:00Z'),
                    EndDateTime=parse('2010-06-06T01:15:00Z'),
                    ReadyDateTime=parse('2010-06-05T23:45:00Z'),
                ),
            ),
            Steps=[
                dict(
                    Name='mr_exciting.woo.20100605.232950.000000: Step 1 of 1',
                    Status=dict(
                        Timeline=dict(
                            StartDateTime=parse('2010-06-05T23:45:00Z'),
                            EndDateTime=parse('2010-06-06T00:15:00Z'),
                        ),
                    ),
                ),
                dict(
                    Name='mr_whatever.meh.20100606.002000.000000: Step 1 of 2',
                    Status=dict(
                        Timeline=dict(
                            StartDateTime=parse('2010-06-06T00:30:00Z'),
                            EndDateTime=parse('2010-06-06T00:45:00Z'),
                        ),
                    ),
                ),
                dict(
                    Name='mr_whatever.meh.20100606.002000.000000: Step 2 of 2',
                    Status=dict(
                        Timeline=dict(
                            StartDateTime=parse('2010-06-06T00:45:00Z'),
                            EndDateTime=parse('2010-06-06T01:00:00Z'),
                        ),
                    ),
                ),
            ],
            **kwargs
        )

        summary = _cluster_to_full_summary(cluster)

        self.assertEqual(summary, {
            'created': utc(2010, 6, 5, 23, 30),
            'end': utc(2010, 6, 6, 1, 15),
            'id': u'j-ISFORJOB',
            'label': u'mr_exciting',
            'name': u'mr_exciting.woo.20100605.232850.000000',
            'nih': 20.0,
            'nih_bbnu': 5.0,
            'nih_billed': 17.5,
            'nih_used': 12.5,
            'num_steps': 3,
            'owner': u'woo',
            'pool': u'reflecting',
            'ran': timedelta(hours=1, minutes=45),
            'ready': utc(2010, 6, 5, 23, 45),
            'state': 'TERMINATED',
            'usage': [{
                # bootstrapping
                'date_to_nih_used': {date(2010, 6, 5): 2.5},
                'date_to_nih_bbnu': {},
                'date_to_nih_billed': {date(2010, 6, 5): 2.5},
                'end': utc(2010, 6, 5, 23, 45),
                'end_billing': utc(2010, 6, 5, 23, 45),
                'hour_to_nih_bbnu': {},
                'hour_to_nih_billed': {utc(2010, 6, 5, 23): 2.5},
                'hour_to_nih_used': {utc(2010, 6, 5, 23): 2.5},
                'label': u'mr_exciting',
                'nih_used': 2.5,
                'nih_bbnu': 0.0,
                'nih_billed': 2.5,
                'owner': u'woo',
                'start': utc(2010, 6, 5, 23, 30),
                'step_num': None,
            }, {
                # mr_exciting, step 1 (and idle time after)
                'date_to_nih_used': {date(2010, 6, 5): 2.5,
                                     date(2010, 6, 6): 2.5},
                'date_to_nih_bbnu': {date(2010, 6, 6): 2.5},
                'date_to_nih_billed': {date(2010, 6, 5): 2.5,
                                       date(2010, 6, 6): 5.0},
                'end': utc(2010, 6, 6, 0, 15),
                'end_billing': utc(2010, 6, 6, 0, 30),
                'hour_to_nih_bbnu': {utc(2010, 6, 6, 0): 2.5},
                'hour_to_nih_billed': {utc(2010, 6, 5, 23): 2.5,
                                       utc(2010, 6, 6, 0): 5.0},
                'hour_to_nih_used': {utc(2010, 6, 5, 23): 2.5,
                                     utc(2010, 6, 6, 0): 2.5},
                'label': u'mr_exciting',
                'nih_used': 5.0,
                'nih_bbnu': 2.5,
                'nih_billed': 7.5,
                'owner': u'woo',
                'start': utc(2010, 6, 5, 23, 45),
                'step_num': 1,
            }, {
                # mr whatever, step 1
                'date_to_nih_used': {date(2010, 6, 6): 2.5},
                'date_to_nih_bbnu': {},
                'date_to_nih_billed': {date(2010, 6, 6): 2.5},
                'end': utc(2010, 6, 6, 0, 45),
                'end_billing': utc(2010, 6, 6, 0, 45),
                'hour_to_nih_bbnu': {},
                'hour_to_nih_billed': {utc(2010, 6, 6, 0): 2.5},
                'hour_to_nih_used': {utc(2010, 6, 6, 0): 2.5},
                'label': u'mr_whatever',
                'nih_used': 2.5,
                'nih_bbnu': 0.0,
                'nih_billed': 2.5,
                'owner': u'meh',
                'start': utc(2010, 6, 6, 0, 30),
                'step_num': 1,
            }, {
                # mr whatever, step 2 (and idle time after)
                'date_to_nih_used': {date(2010, 6, 6): 2.5},
                'date_to_nih_bbnu': {date(2010, 6, 6): 2.5},
                'date_to_nih_billed': {date(2010, 6, 6): 5.0},
                'end': utc(2010, 6, 6, 1, 0),
                'end_billing': utc(2010, 6, 6, 1, 15),
                'hour_to_nih_bbnu': {utc(2010, 6, 6, 1): 2.5},
                'hour_to_nih_billed': {utc(2010, 6, 6, 0): 2.5,
                                       utc(2010, 6, 6, 1): 2.5},
                'hour_to_nih_used': {utc(2010, 6, 6, 0): 2.5},
                'label': u'mr_whatever',
                'nih_used': 2.5,
                'nih_bbnu': 2.5,
                'nih_billed': 5.0,
                'owner': u'meh',
                'start': utc(2010, 6, 6, 0, 45),
                'step_num': 2,
            }],
        })


class SubdivideIntervalByDateTestCase(BasicTestCase):

    def test_zero_interval(self):
        self.assertEqual(
            _subdivide_interval_by_date(
                utc(2010, 6, 6, 4, 26),
                utc(2010, 6, 6, 4, 26),
            ),
            {}
        )

    def test_same_day(self):
        self.assertEqual(
            _subdivide_interval_by_date(
                utc(2010, 6, 6, 4, 0),
                utc(2010, 6, 6, 6, 0),
            ),
            {date(2010, 6, 6): 7200.0}
        )

    def test_start_at_midnight(self):
        self.assertEqual(
            _subdivide_interval_by_date(
                utc(2010, 6, 6, 0, 0),
                utc(2010, 6, 6, 5, 0),
            ),
            {date(2010, 6, 6): 18000.0}
        )

    def test_end_at_midnight(self):
        self.assertEqual(
            _subdivide_interval_by_date(
                utc(2010, 6, 5, 23, 0),
                utc(2010, 6, 6, 0, 0),
            ),
            {date(2010, 6, 5): 3600.0}
        )

    def test_split_over_midnight(self):
        self.assertEqual(
            _subdivide_interval_by_date(
                utc(2010, 6, 5, 23, 0),
                utc(2010, 6, 6, 5, 0),
            ),
            {date(2010, 6, 5): 3600.0,
             date(2010, 6, 6): 18000.0}
        )

    def test_full_days(self):
        self.assertEqual(
            _subdivide_interval_by_date(
                utc(2010, 6, 5, 23, 0),
                utc(2010, 6, 10, 5, 0),
            ),
            {date(2010, 6, 5): 3600.0,
             date(2010, 6, 6): 86400.0,
             date(2010, 6, 7): 86400.0,
             date(2010, 6, 8): 86400.0,
             date(2010, 6, 9): 86400.0,
             date(2010, 6, 10): 18000.0}
        )


class SubdivideIntervalByHourTestCase(BasicTestCase):

    def test_zero_interval(self):
        self.assertEqual(
            _subdivide_interval_by_hour(
                utc(2010, 6, 6, 4, 26),
                utc(2010, 6, 6, 4, 26),
            ),
            {}
        )

    def test_same_hour(self):
        self.assertEqual(
            _subdivide_interval_by_hour(
                utc(2010, 6, 6, 4, 24),
                utc(2010, 6, 6, 4, 26),
            ),
            {utc(2010, 6, 6, 4): 120.0}
        )

    def test_start_at_midnight(self):
        self.assertEqual(
            _subdivide_interval_by_hour(
                utc(2010, 6, 6, 0, 0),
                utc(2010, 6, 6, 0, 3),
            ),
            {utc(2010, 6, 6, 0): 180.0}
        )

    def test_end_at_midnight(self):
        self.assertEqual(
            _subdivide_interval_by_hour(
                utc(2010, 6, 5, 23, 55),
                utc(2010, 6, 6, 0, 0),
            ),
            {utc(2010, 6, 5, 23): 300.0}
        )

    def test_split_over_midnight(self):
        self.assertEqual(
            _subdivide_interval_by_hour(
                utc(2010, 6, 5, 23, 55),
                utc(2010, 6, 6, 0, 3),
            ),
            {utc(2010, 6, 5, 23): 300.0,
             utc(2010, 6, 6, 0): 180.0}
        )

    def test_full_hours(self):
        self.assertEqual(
            _subdivide_interval_by_hour(
                utc(2010, 6, 5, 23, 40),
                utc(2010, 6, 6, 2, 10),
            ),
            {utc(2010, 6, 5, 23): 1200.0,
             utc(2010, 6, 6, 0): 3600.0,
             utc(2010, 6, 6, 1): 3600.0,
             utc(2010, 6, 6, 2): 600.0}
        )


class PercentTestCase(BasicTestCase):

    def test_basic(self):
        self.assertEqual(62.5, _percent(5, 8))

    def test_default(self):
        self.assertEqual(0.0, _percent(1, 0))
        self.assertEqual(0.0, _percent(0, 0))
        self.assertEqual(None, _percent(0, 0, default=None))



# Copyright 2009-2013 Yelp and Contributors
# Copyright 2015 Yelp
# Copyright 2017 Yelp and Contributors
# Copyright 2018 Yelp
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import bz2
import socket
from ssl import SSLError

from botocore.exceptions import ClientError

from mrjob.fs.s3 import S3Filesystem
from mrjob.fs.s3 import _wrap_aws_client
from mrjob.fs.s3 import _AWS_MAX_TRIES

from tests.compress import gzip_compress
from tests.mock_boto3 import MockBoto3TestCase
from tests.py2 import patch


class CatTestCase(MockBoto3TestCase):

    def setUp(self):
        super(CatTestCase, self).setUp()
        self.fs = S3Filesystem()

    def test_cat_uncompressed(self):
        self.add_mock_s3_data(
            {'walrus': {'data/foo': b'foo\nfoo\n'}})

        self.assertEqual(
            b''.join(self.fs._cat_file('s3://walrus/data/foo')),
            b'foo\nfoo\n')

    def test_cat_bz2(self):
        self.add_mock_s3_data(
            {'walrus': {'data/foo.bz2': bz2.compress(b'foo\n' * 1000)}})

        self.assertEqual(
            b''.join(self.fs._cat_file('s3://walrus/data/foo.bz2')),
            b'foo\n' * 1000)

    def test_cat_gz(self):
        self.add_mock_s3_data(
            {'walrus': {'data/foo.gz': gzip_compress(b'foo\n' * 10000)}})

        self.assertEqual(
            b''.join(self.fs._cat_file('s3://walrus/data/foo.gz')),
            b'foo\n' * 10000)

    def test_chunks_file(self):
        self.add_mock_s3_data(
            {'walrus': {'data/foo': b'foo\n' * 1000}})

        self.assertGreater(
            len(list(self.fs._cat_file('s3://walrus/data/foo'))),
            1)


class S3FSTestCase(MockBoto3TestCase):

    def setUp(self):
        super(S3FSTestCase, self).setUp()
        self.fs = S3Filesystem()

    def test_ls_key(self):
        self.add_mock_s3_data(
            {'walrus': {'data/foo': b''}})

        self.assertEqual(list(self.fs.ls('s3://walrus/data/foo')),
                         ['s3://walrus/data/foo'])

    def test_ls_recursively(self):
        self.add_mock_s3_data(
            {'walrus': {'data/bar': b'',
                        'data/bar/baz': b'',
                        'data/foo': b'',
                        'qux': b''}})

        uris = [
            's3://walrus/data/bar',
            's3://walrus/data/bar/baz',
            's3://walrus/data/foo',
            's3://walrus/qux',
        ]

        self.assertEqual(list(self.fs.ls('s3://walrus/')), uris)
        self.assertEqual(list(self.fs.ls('s3://walrus/*')), uris)

        self.assertEqual(list(self.fs.ls('s3://walrus/data')), uris[:-1])
        self.assertEqual(list(self.fs.ls('s3://walrus/data/')), uris[:-1])
        self.assertEqual(list(self.fs.ls('s3://walrus/data/*')), uris[:-1])

    def test_ls_globs(self):
        self.add_mock_s3_data(
            {'w': {'a': b'',
                   'a/b': b'',
                   'ab': b'',
                   'b': b''}})

        self.assertEqual(list(self.fs.ls('s3://w/')),
                         ['s3://w/a', 's3://w/a/b', 's3://w/ab', 's3://w/b'])
        self.assertEqual(list(self.fs.ls('s3://w/*')),
                         ['s3://w/a', 's3://w/a/b', 's3://w/ab', 's3://w/b'])
        self.assertEqual(list(self.fs.ls('s3://w/*/')),
                         ['s3://w/a/b'])
        self.assertEqual(list(self.fs.ls('s3://w/*/*')),
                         ['s3://w/a/b'])
        self.assertEqual(list(self.fs.ls('s3://w/a?')),
                         ['s3://w/ab'])
        # * can match /
        self.assertEqual(list(self.fs.ls('s3://w/a*')),
                         ['s3://w/a', 's3://w/a/b', 's3://w/ab'])
        self.assertEqual(list(self.fs.ls('s3://w/*b')),
                         ['s3://w/a/b', 's3://w/ab', 's3://w/b'])

    def test_ls_s3n(self):
        self.add_mock_s3_data(
            {'walrus': {'data/bar': b'abc123',
                        'data/baz': b'123abc'}})

        self.assertEqual(list(self.fs.ls('s3n://walrus/data/*')),
                         ['s3n://walrus/data/bar',
                          's3n://walrus/data/baz'])

    def test_ls_s3a(self):
        self.add_mock_s3_data(
            {'walrus': {'data/bar': b'abc123',
                        'data/baz': b'123abc'}})

        self.assertEqual(list(self.fs.ls('s3a://walrus/data/*')),
                         ['s3a://walrus/data/bar',
                          's3a://walrus/data/baz'])

    def test_du(self):
        self.add_mock_s3_data({
            'walrus': {'data/foo': b'abcde',
                       'data/bar/baz': b'fgh'}})

        self.assertEqual(self.fs.du('s3://walrus/'), 8)
        self.assertEqual(self.fs.du('s3://walrus/data/foo'), 5)
        self.assertEqual(self.fs.du('s3://walrus/data/bar/baz'), 3)

    def test_exists(self):
        self.add_mock_s3_data({
            'walrus': {'data/foo': b'abcd'}})
        self.assertEqual(self.fs.exists('s3://walrus/data/foo'), True)
        self.assertEqual(self.fs.exists('s3://walrus/data/bar'), False)

    def test_rm(self):
        self.add_mock_s3_data({
            'walrus': {'foo': b''}})

        self.assertEqual(self.fs.exists('s3://walrus/foo'), True)
        self.fs.rm('s3://walrus/foo')
        self.assertEqual(self.fs.exists('s3://walrus/foo'), False)

    def test_rm_dir(self):
        self.add_mock_s3_data({
            'walrus': {'data/foo': b'',
                       'data/bar/baz': b''}})

        self.assertEqual(self.fs.exists('s3://walrus/data/foo'), True)
        self.assertEqual(self.fs.exists('s3://walrus/data/bar/baz'), True)
        self.fs.rm('s3://walrus/data')
        self.assertEqual(self.fs.exists('s3://walrus/data/foo'), False)
        self.assertEqual(self.fs.exists('s3://walrus/data/bar/baz'), False)

    def test_md5sum(self):
        self.add_mock_s3_data({
            'walrus': {'data/foo': b'abcd'}})

        self.assertEqual(self.fs.md5sum('s3://walrus/data/foo'),
                         'e2fc714c4727ee9395f324cd2e7f331f')

    def test_touchz(self):
        self.add_mock_s3_data({'walrus': {}})

        self.assertEqual(list(self.fs.ls('s3://walrus/')), [])

        self.fs.touchz('s3://walrus/empty')

        self.assertEqual(list(self.fs.ls('s3://walrus/')),
                         ['s3://walrus/empty'])

    def test_okay_to_touchz_empty_file(self):
        self.add_mock_s3_data({'walrus': {'empty': b''}})

        self.assertEqual(list(self.fs.ls('s3://walrus/')),
                         ['s3://walrus/empty'])

        self.fs.touchz('s3://walrus/empty')

        self.assertEqual(list(self.fs.ls('s3://walrus/')),
                         ['s3://walrus/empty'])

    def test_cant_touchz_file_with_contents(self):
        self.add_mock_s3_data({'walrus': {'full': b'contents'}})

        self.assertRaises(OSError,
                          self.fs.touchz, 's3://walrus/full')

    def test_mkdir_does_nothing(self):
        self.add_mock_s3_data({'walrus': {}})

        self.assertEqual(list(self.fs.ls('s3://walrus/')), [])

        self.fs.mkdir('s3://walrus/data')

        self.assertEqual(list(self.fs.ls('s3://walrus/')), [])

    # S3-specific utilities

    def test_get_all_bucket_names(self):
        self.add_mock_s3_data({'walrus': {}, 'kitteh': {}})

        self.assertEqual(list(self.fs.get_all_bucket_names()),
                         ['kitteh', 'walrus'])


class S3FSRegionTestCase(MockBoto3TestCase):

    def test_default_endpoint(self):
        fs = S3Filesystem()

        client = fs.make_s3_client()
        self.assertEqual(client.meta.endpoint_url,
                         'https://s3.amazonaws.com')

        resource = fs.make_s3_resource()
        self.assertEqual(resource.meta.client.meta.endpoint_url,
                         'https://s3.amazonaws.com')

    def test_force_s3_endpoint_host(self):
        fs = S3Filesystem(s3_endpoint='myproxy')

        client = fs.make_s3_client()
        self.assertEqual(client.meta.endpoint_url,
                         'https://myproxy')

        resource = fs.make_s3_resource()
        self.assertEqual(resource.meta.client.meta.endpoint_url,
                         'https://myproxy')

    def test_force_s3_endpoint_url(self):
        fs = S3Filesystem(s3_endpoint='https://myproxy:8080')

        client = fs.make_s3_client()
        self.assertEqual(client.meta.endpoint_url,
                         'https://myproxy:8080')

        resource = fs.make_s3_resource()
        self.assertEqual(resource.meta.client.meta.endpoint_url,
                         'https://myproxy:8080')

    def test_force_s3_endpoint_region(self):
        # this is the actual mrjob default region
        fs = S3Filesystem(s3_region='us-west-2')

        client = fs.make_s3_client()
        self.assertEqual(client.meta.endpoint_url,
                         'https://s3-us-west-2.amazonaws.com')
        self.assertEqual(client.meta.region_name,
                         'us-west-2')

        resource = fs.make_s3_resource()
        self.assertEqual(resource.meta.client.meta.endpoint_url,
                         'https://s3-us-west-2.amazonaws.com')
        self.assertEqual(resource.meta.client.meta.region_name,
                         'us-west-2')

    def test_endpoint_for_bucket_in_us_west_1(self):
        self.add_mock_s3_data({'walrus': {}}, location='us-west-1')

        fs = S3Filesystem()

        bucket = fs.get_bucket('walrus')
        self.assertEqual(bucket.meta.client.meta.region_name,
                         'us-west-1')

    def test_get_location_is_forbidden(self):
        self.add_mock_s3_data({'walrus': {}}, location='us-west-2')

        fs = S3Filesystem()

        access_denied_error = ClientError(
            dict(
                Error=dict(
                    Code='AccessDenied',
                    Message='Access Denied',
                ),
                ResponseMetadata=dict(
                    HTTPStatusCode=403
                ),
            ),
            'GetBucketLocation')

        with patch(
                'tests.mock_boto3.s3.MockS3Client.get_bucket_location',
                side_effect=access_denied_error):

            bucket = fs.get_bucket('walrus')

        self.assertEqual(bucket.meta.client.meta.endpoint_url,
                         'https://s3.amazonaws.com')
        self.assertEqual(bucket.meta.client.meta.region_name, 'us-east-1')

    def test_bucket_does_not_exist(self):
        fs = S3Filesystem()

        self.assertRaises(ClientError, fs.get_bucket, 'walrus')

    def test_endpoint_for_bucket_in_us_east_1(self):
        # location constraint for us-east-1 is '', not 'us-east-1'
        self.add_mock_s3_data({'walrus': {}}, location='')

        fs = S3Filesystem()

        bucket = fs.get_bucket('walrus')
        self.assertEqual(bucket.meta.client.meta.endpoint_url,
                         'https://s3.amazonaws.com')

    def test_buckets_from_forced_s3_endpoint(self):
        self.add_mock_s3_data({'walrus-east': {}}, location='us-east-2')

        fs = S3Filesystem(s3_endpoint='s3-us-west-2.amazonaws.com')

        bucket_east = fs.get_bucket('walrus-east')

        with patch('tests.mock_boto3.s3.MockS3Client.get_bucket_location'
                   ) as mock_gbl:
            # won't actually be able to access this bucket from this endpoint,
            # but boto3 doesn't check that on bucket creation
            self.assertEqual(bucket_east.meta.client.meta.endpoint_url,
                             'https://s3-us-west-2.amazonaws.com')
            # no reason to check bucket location if endpoint is forced
            self.assertFalse(mock_gbl.called)

    def test_create_bucket_with_no_region(self):
        fs = S3Filesystem()

        fs.create_bucket('walrus')

        s3_client = fs.make_s3_client()
        self.assertEqual(
            s3_client.get_bucket_location('walrus')['LocationConstraint'],
            None)

        bucket = fs.get_bucket('walrus')

        self.assertEqual(bucket.meta.client.meta.endpoint_url,
                         'https://s3.amazonaws.com')
        self.assertEqual(bucket.meta.client.meta.region_name,
                         'us-east-1')

    def test_create_bucket_in_us_east_1(self):
        fs = S3Filesystem()

        fs.create_bucket('walrus', region='us-east-1')

        s3_client = fs.make_s3_client()
        self.assertEqual(
            s3_client.get_bucket_location('walrus')['LocationConstraint'],
            None)

        bucket = fs.get_bucket('walrus')

        self.assertEqual(bucket.meta.client.meta.endpoint_url,
                         'https://s3.amazonaws.com')
        self.assertEqual(bucket.meta.client.meta.region_name,
                         'us-east-1')

    def test_create_bucket_in_us_west_2(self):
        fs = S3Filesystem()

        fs.create_bucket('walrus', region='us-west-2')

        s3_client = fs.make_s3_client()
        self.assertEqual(
            s3_client.get_bucket_location('walrus')['LocationConstraint'],
            'us-west-2')

        bucket = fs.get_bucket('walrus')

        self.assertEqual(bucket.meta.client.meta.endpoint_url,
                         'https://s3-us-west-2.amazonaws.com')
        self.assertEqual(bucket.meta.client.meta.region_name,
                         'us-west-2')


class WrapAWSClientTestCase(MockBoto3TestCase):
    """Test that _wrap_aws_client() works as expected.

    We're going to wrap S3Client.list_buckets(), but it should work the
    same on any method on any AWS client/resource."""

    def setUp(self):
        super(WrapAWSClientTestCase, self).setUp()

        # don't actually wait between retries
        self.sleep = self.start(patch('time.sleep'))

        self.log = self.start(patch('mrjob.retry.log'))

        self.list_buckets = self.start(patch(
            'tests.mock_boto3.s3.MockS3Client.list_buckets',
            side_effect=[dict(Buckets=[])]))

        self.client = self.client('s3')
        self.wrapped_client = _wrap_aws_client(self.client)

    def add_transient_error(self, ex):
        self.list_buckets.side_effect = (
            [ex] + list(self.list_buckets.side_effect))

    def test_unwrapped_no_errors(self):
        # just a sanity check of our mocks
        self.assertEqual(self.client.list_buckets(), dict(Buckets=[]))

    def test_unwrapped_with_errors(self):
        self.add_transient_error(socket.error(104, 'Connection reset by peer'))

        self.assertRaises(socket.error, self.client.list_buckets)

    def test_wrapped_no_errors(self):
        self.assertEqual(self.wrapped_client.list_buckets(), dict(Buckets=[]))

        self.assertFalse(self.log.info.called)

    def assert_retry(self, ex):
        self.add_transient_error(ex)
        self.assertEqual(self.wrapped_client.list_buckets(), dict(Buckets=[]))
        self.assertTrue(self.log.info.called)

    def assert_no_retry(self, ex):
        self.add_transient_error(ex)
        self.assertRaises(ex.__class__, self.wrapped_client.list_buckets)
        self.assertFalse(self.log.info.called)

    def test_socket_connection_reset_by_peer(self):
        self.assert_retry(socket.error(104, 'Connection reset by peer'))

    def test_socket_connection_timed_out(self):
        self.assert_retry(socket.error(110, 'Connection timed out'))

    def test_other_socket_errors(self):
        self.assert_no_retry(socket.error(111, 'Connection refused'))

    def test_ssl_read_op_timed_out_error(self):
        self.assert_retry(SSLError('The read operation timed out'))

    def test_ssl_write_op_timed_out_error(self):
        self.assert_retry(SSLError('The write operation timed out'))

    def test_other_ssl_error(self):
        self.assert_no_retry(SSLError('certificate verify failed'))

    def test_throttling_error(self):
        self.assert_retry(ClientError(
            dict(
                Error=dict(
                    Code='ThrottlingError'
                )
            ),
            'ListBuckets'))

    def test_aws_505_error(self):
        self.assert_retry(ClientError(
            dict(
                ResponseMetadata=dict(
                    HTTPStatusCode=505
                )
            ),
            'ListBuckets'))

    def test_other_client_error(self):
        self.assert_no_retry(ClientError(
            dict(
                Error=dict(
                    Code='AccessDenied',
                    Message='Access Denied',
                ),
                ResponseMetadata=dict(
                    HTTPStatusCode=403
                ),
            ),
            'GetBucketLocation'))

    def test_other_error(self):
        self.assert_no_retry(ValueError())

    def test_two_retriable_errors(self):
        self.add_transient_error(socket.error(110, 'Connection timed out'))
        self.add_transient_error(SSLError('The read operation timed out'))
        self.assertEqual(self.wrapped_client.list_buckets(), dict(Buckets=[]))
        self.assertTrue(self.log.info.called)

    def test_retriable_and_no_retriable_error(self):
        # errors are prepended to side effects
        # we want socket.error to happen first
        self.add_transient_error(ValueError())
        self.add_transient_error(socket.error(110, 'Connection timed out'))

        self.assertRaises(ValueError, self.wrapped_client.list_buckets)
        self.assertTrue(self.log.info.called)

    def test_eventually_give_up(self):
        for _ in range(_AWS_MAX_TRIES + 1):
            self.add_transient_error(socket.error(110, 'Connection timed out'))

        self.assertRaises(socket.error, self.wrapped_client.list_buckets)
        self.assertTrue(self.log.info.called)

    def test_min_backoff(self):
        self.wrapped_client = _wrap_aws_client(self.client, min_backoff=1000)

        self.add_transient_error(socket.error(110, 'Connection timed out'))

        self.assertEqual(self.wrapped_client.list_buckets(), dict(Buckets=[]))

        self.sleep.assert_called_with(1000)

        self.assertTrue(self.log.info.called)



# Copyright 2009-2013 Yelp and Contributors
# Copyright 2015 Yelp
# Copyright 2017 Yelp
# Copyright 2018 Yelp
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Make sure all of our protocols work as advertised."""
import unittest
from tests.sandbox import BasicTestCase
from unittest import skipIf

from mrjob.protocol import BytesProtocol
from mrjob.protocol import BytesValueProtocol
from mrjob.protocol import JSONProtocol
from mrjob.protocol import JSONValueProtocol
from mrjob.protocol import PickleProtocol
from mrjob.protocol import PickleValueProtocol
from mrjob.protocol import RapidJSONProtocol
from mrjob.protocol import RapidJSONValueProtocol
from mrjob.protocol import RawProtocol
from mrjob.protocol import RawValueProtocol
from mrjob.protocol import ReprProtocol
from mrjob.protocol import ReprValueProtocol
from mrjob.protocol import SimpleJSONProtocol
from mrjob.protocol import SimpleJSONValueProtocol
from mrjob.protocol import StandardJSONProtocol
from mrjob.protocol import StandardJSONValueProtocol
from mrjob.protocol import TextProtocol
from mrjob.protocol import TextValueProtocol
from mrjob.protocol import UltraJSONProtocol
from mrjob.protocol import UltraJSONValueProtocol
from mrjob.protocol import rapidjson
from mrjob.protocol import simplejson
from mrjob.protocol import ujson
from mrjob.py2 import PY2


class Point(object):
    """A simple class to test encoding of objects."""

    def __init__(self, x, y):
        self.x = x
        self.y = y

    def __repr__(self):
        return '%s.%s(%r, %r)' % (self.__module__, self.__class__.__name__,
                                  self.x, self.y)

    def __eq__(self, other):
        if not isinstance(other, Point):
            return False

        return (self.x, self.y) == (other.x, other.y)

    def __ne__(self, other):
        return not self.__eq__(other)


# keys and values that JSON protocols should encode/decode correctly
JSON_KEYS_AND_VALUES = [
    (None, None),
    (1, 2),
    ('foo', 'bar'),
    ([1, 2, 3], []),
    ({'apples': 5}, {'oranges': 20}),
    (u'Qu\xe9bec', u'Ph\u1ede'),
    ('\t', '\n'),
]

# keys and values that repr protocols should encode/decode correctly
REPR_KEYS_AND_VALUES = JSON_KEYS_AND_VALUES + [
    ((1, 2), (3, 4)),
    (b'0\xa2', b'\xe9'),
    (set([1]), set()),
]

# keys and values that pickle protocols should encode/decode properly
PICKLE_KEYS_AND_VALUES = REPR_KEYS_AND_VALUES + [
    (Point(2, 3), Point(1, 4)),
]


class ProtocolTestCase(BasicTestCase):

    def assertRoundTripOK(self, protocol, key, value):
        """Assert that we can encode and decode the given key and value,
        and get the same key and value we started with."""
        self.assertEqual((key, value),
                         protocol.read(protocol.write(key, value)))

    def assertRoundTripWithTrailingTabOK(self, protocol, key, value):
        """Assert that we can encode the given key and value, add a
        trailing tab (which Hadoop sometimes does), and decode it
        to get the same key and value we started with."""
        self.assertEqual((key, value),
                         protocol.read(protocol.write(key, value) + b'\t'))

    def assertCantEncode(self, protocol, key, value):
        self.assertRaises(Exception, protocol.write, key, value)

    def assertCantDecode(self, protocol, data):
        self.assertRaises(Exception, protocol.read, data)


class JSONProtocolAliasesTestCase(BasicTestCase):

    def test_use_ujson_or_simplejson_if_installed(self):
        # these aliases are determined at compile time, so there
        # isn't a straightforward way to test this through patches
        if ujson:
            self.assertEqual(JSONProtocol, UltraJSONProtocol)
            self.assertEqual(JSONValueProtocol, UltraJSONValueProtocol)
        elif rapidjson and not PY2:
            self.assertEqual(JSONProtocol, RapidJSONProtocol)
            self.assertEqual(JSONValueProtocol, RapidJSONValueProtocol)
        elif simplejson:
            self.assertEqual(JSONProtocol, SimpleJSONProtocol)
            self.assertEqual(JSONValueProtocol, SimpleJSONValueProtocol)
        else:
            self.assertEqual(JSONProtocol, StandardJSONProtocol)
            self.assertEqual(JSONValueProtocol, StandardJSONValueProtocol)


class StandardJSONProtocolTestCase(ProtocolTestCase):

    PROTOCOL = StandardJSONProtocol()

    def test_round_trip(self):
        for k, v in JSON_KEYS_AND_VALUES:
            self.assertRoundTripOK(self.PROTOCOL, k, v)

    def test_round_trip_with_trailing_tab(self):
        for k, v in JSON_KEYS_AND_VALUES:
            self.assertRoundTripWithTrailingTabOK(self.PROTOCOL, k, v)

    def test_uses_json_format(self):
        KEY = ['a', 1]
        VALUE = {'foo': 'bar'}
        ENCODED = b'["a", 1]\t{"foo": "bar"}'

        self.assertEqual((KEY, VALUE), self.PROTOCOL.read(ENCODED))
        # ujson and rapidjson don't use spaces
        self.assertEqual(self.PROTOCOL.write(KEY, VALUE).replace(b' ', b''),
                         ENCODED.replace(b' ', b''))

    def test_tuples_become_lists(self):
        # JSON should convert tuples into lists
        self.assertEqual(
            ([1, 2], [3, 4]),
            self.PROTOCOL.read(self.PROTOCOL.write((1, 2), (3, 4))))

    def test_numerical_keys_become_strs(self):
        # JSON should convert numbers to strings when they are dict keys
        self.assertEqual(
            ({'1': 2}, {'3': 4}),
            self.PROTOCOL.read(self.PROTOCOL.write({1: 2}, {3: 4})))

    def test_bad_data(self):
        self.assertCantDecode(self.PROTOCOL, b'{@#$@#!^&*$%^')

    def test_bad_keys_and_values(self):
        # only unicodes (or bytes in utf-8) are allowed
        self.assertCantEncode(self.PROTOCOL, b'0\xa2', b'\xe9')

        # dictionaries have to have strings as keys
        self.assertCantEncode(self.PROTOCOL, {(1, 2): 3}, None)

        # sets don't exist in JSON
        self.assertCantEncode(self.PROTOCOL, set([1]), set())

        # Point class has no representation in JSON
        self.assertCantEncode(self.PROTOCOL, Point(2, 3), Point(1, 4))


@skipIf(simplejson is None, 'simplejson module not installed')
class SimpleJSONProtocolTestCase(StandardJSONProtocolTestCase):

    PROTOCOL = SimpleJSONProtocol()


@skipIf(rapidjson is None, 'rapidjson module not installed')
class RapidJSONProtocolTestCase(StandardJSONProtocolTestCase):

    PROTOCOL = RapidJSONProtocol()

    @unittest.skip('rapidjson only allows strings as keys')
    def test_numerical_keys_become_strs(self):
        pass

    def test_bad_keys_and_values(self):
        # dictionaries have to have strings as keys
        self.assertCantEncode(self.PROTOCOL, {(1, 2): 3}, None)

        # sets don't exist in JSON
        self.assertCantEncode(self.PROTOCOL, set([1]), set())

        # Point class has no representation in JSON
        self.assertCantEncode(self.PROTOCOL, Point(2, 3), Point(1, 4))


@skipIf(ujson is None, 'ujson module not installed')
class UltraJSONProtocolTestCase(StandardJSONProtocolTestCase):

    PROTOCOL = UltraJSONProtocol()

    def test_bad_keys_and_values(self):
        # seems like the only thing ujson won't encode is non-UTF-8 bytes
        self.assertCantEncode(self.PROTOCOL, b'0\xa2', b'\xe9')


class StandardJSONValueProtocolTestCase(ProtocolTestCase):

    PROTOCOL = StandardJSONValueProtocol()

    def test_round_trip(self):
        for _, v in JSON_KEYS_AND_VALUES:
            self.assertRoundTripOK(self.PROTOCOL, None, v)

    def test_round_trip_with_trailing_tab(self):
        for _, v in JSON_KEYS_AND_VALUES:
            self.assertRoundTripWithTrailingTabOK(self.PROTOCOL, None, v)

    def test_uses_json_format(self):
        VALUE = {'foo': 'bar'}
        ENCODED = b'{"foo": "bar"}'

        self.assertEqual((None, VALUE), self.PROTOCOL.read(ENCODED))
        self.assertEqual(self.PROTOCOL.write(None, VALUE).replace(b' ', b''),
                         ENCODED.replace(b' ', b''))

    def test_tuples_become_lists(self):
        # JSON should convert tuples into lists
        self.assertEqual(
            (None, [3, 4]),
            self.PROTOCOL.read(self.PROTOCOL.write(None, (3, 4))))

    def test_numerical_keys_become_strs(self):
        # JSON should convert numbers to strings when they are dict keys
        self.assertEqual(
            (None, {'3': 4}),
            self.PROTOCOL.read(self.PROTOCOL.write(None, {3: 4})))

    def test_bad_data(self):
        self.assertCantDecode(self.PROTOCOL, b'{@#$@#!^&*$%^')

    def test_bad_keys_and_values(self):
        # seems like the only thing ujson won't encode is non-UTF-8 bytes
        self.assertCantEncode(self.PROTOCOL, None, b'\xe9')

        # dictionaries have to have strings as keys
        self.assertCantEncode(self.PROTOCOL, None, {(1, 2): 3})

        # sets don't exist in JSON
        self.assertCantEncode(self.PROTOCOL, None, set())

        # Point class has no representation in JSON
        self.assertCantEncode(self.PROTOCOL, None, Point(1, 4))


@skipIf(simplejson is None, 'simplejson module not installed')
class SimpleJSONValueProtocolTestCase(StandardJSONValueProtocolTestCase):

    PROTOCOL = SimpleJSONValueProtocol()


@skipIf(rapidjson is None, 'rapidjson module not installed')
class RapidJSONValueProtocolTestCase(StandardJSONValueProtocolTestCase):

    PROTOCOL = RapidJSONValueProtocol()

    @unittest.skip('rapidjson only allows strings as keys')
    def test_numerical_keys_become_strs(self):
        pass

    def test_bad_keys_and_values(self):
        # dictionaries have to have strings as keys
        self.assertCantEncode(self.PROTOCOL, None, {(1, 2): 3})

        # sets don't exist in JSON
        self.assertCantEncode(self.PROTOCOL, None, set())

        # Point class has no representation in JSON
        self.assertCantEncode(self.PROTOCOL, None, Point(1, 4))


@skipIf(ujson is None, 'ujson module not installed')
class UltraJSONValueProtocolTestCase(StandardJSONValueProtocolTestCase):

    PROTOCOL = UltraJSONValueProtocol()

    def test_uses_json_format(self):
        VALUE = {'foo': 'bar'}
        ENCODED = b'{"foo":"bar"}'  # no whitespace in ujson

        self.assertEqual((None, VALUE), self.PROTOCOL.read(ENCODED))
        self.assertEqual(self.PROTOCOL.write(None, VALUE), ENCODED)

    def test_bad_keys_and_values(self):
        # seems like the only thing ujson won't encode is non-UTF-8 bytes
        self.assertCantEncode(self.PROTOCOL, None, b'\xe9')


class PickleProtocolTestCase(ProtocolTestCase):

    def test_round_trip(self):
        for k, v in PICKLE_KEYS_AND_VALUES:
            self.assertRoundTripOK(PickleProtocol(), k, v)

    def test_round_trip_with_trailing_tab(self):
        for k, v in PICKLE_KEYS_AND_VALUES:
            self.assertRoundTripWithTrailingTabOK(PickleProtocol(), k, v)

    def test_bad_data(self):
        self.assertCantDecode(PickleProtocol(), b'{@#$@#!^&*$%^')

    # no tests of what encoded data looks like; pickle is an opaque protocol


class PickleValueProtocolTestCase(ProtocolTestCase):

    def test_round_trip(self):
        for _, v in PICKLE_KEYS_AND_VALUES:
            self.assertRoundTripOK(PickleValueProtocol(), None, v)

    def test_round_trip_with_trailing_tab(self):
        for _, v in PICKLE_KEYS_AND_VALUES:
            self.assertRoundTripWithTrailingTabOK(
                PickleValueProtocol(), None, v)

    def test_bad_data(self):
        self.assertCantDecode(PickleValueProtocol(), b'{@#$@#!^&*$%^')

    # no tests of what encoded data looks like; pickle is an opaque protocol


class RawProtocolAliasesTestCase(BasicTestCase):

    def test_raw_protocol_aliases(self):
        if PY2:
            self.assertEqual(RawProtocol, BytesProtocol)
            self.assertEqual(RawValueProtocol, BytesValueProtocol)
        else:
            self.assertEqual(RawProtocol, TextProtocol)
            self.assertEqual(RawValueProtocol, TextValueProtocol)


class BytesValueProtocolTestCase(ProtocolTestCase):

    def test_dumps_keys(self):
        self.assertEqual(BytesValueProtocol().write(b'foo', b'bar'), b'bar')

    def test_reads_raw_line(self):
        self.assertEqual(BytesValueProtocol().read(b'foobar'),
                         (None, b'foobar'))

    def test_bytestrings(self):
        self.assertRoundTripOK(BytesValueProtocol(), None, b'\xe90\c1a')

    def test_no_strip(self):
        self.assertEqual(BytesValueProtocol().read(b'foo\t \n\n'),
                         (None, b'foo\t \n\n'))


class TextValueProtocolTestCase(ProtocolTestCase):

    def test_dumps_keys(self):
        self.assertEqual(TextValueProtocol().write(u'foo', u'bar'), b'bar')

    def test_converts_raw_line_to_unicode(self):
        self.assertEqual(TextValueProtocol().read(b'foobar'),
                         (None, u'foobar'))

    def test_utf_8_decode(self):
        self.assertEqual(TextValueProtocol().read(b'caf\xc3\xa9'),
                         (None, u'caf\xe9'))

    def test_fall_back_to_latin_1(self):
        self.assertEqual(TextValueProtocol().read(b'caf\xe9'),
                         (None, u'caf\xe9'))

    def test_no_strip(self):
        self.assertEqual(TextValueProtocol().read(b'foo\t \n\n'),
                         (None, u'foo\t \n\n'))


class BytesProtocolTestCase(ProtocolTestCase):

    def test_round_trip(self):
        self.assertRoundTripOK(BytesProtocol(), b'foo', b'bar')
        self.assertRoundTripOK(BytesProtocol(), b'foo', None)
        self.assertRoundTripOK(BytesProtocol(), b'foo', b'')
        self.assertRoundTripOK(BytesProtocol(), b'caf\xe9', b'\xe90\c1a')

    def test_no_tabs(self):
        self.assertEqual(BytesProtocol().write(b'foo', None), b'foo')
        self.assertEqual(BytesProtocol().write(None, b'foo'), b'foo')
        self.assertEqual(BytesProtocol().read(b'foo'), (b'foo', None))

        self.assertEqual(BytesProtocol().write(b'', None), b'')
        self.assertEqual(BytesProtocol().write(None, None), b'')
        self.assertEqual(BytesProtocol().write(None, b''), b'')
        self.assertEqual(BytesProtocol().read(b''), (b'', None))

    def test_extra_tabs(self):
        self.assertEqual(BytesProtocol().write(b'foo', b'bar\tbaz'),
                         b'foo\tbar\tbaz')
        self.assertEqual(BytesProtocol().write(b'foo\tbar', b'baz'),
                         b'foo\tbar\tbaz')
        self.assertEqual(BytesProtocol().read(b'foo\tbar\tbaz'),
                         (b'foo', b'bar\tbaz'))

    def test_no_strip(self):
        self.assertEqual(BytesProtocol().read(b'foo\t \n\n'),
                         (b'foo', b' \n\n'))


class TextProtocolTestCase(ProtocolTestCase):

    def test_round_trip(self):
        self.assertRoundTripOK(TextProtocol(), u'foo', u'bar')
        self.assertRoundTripOK(TextProtocol(), u'foo', None)
        self.assertRoundTripOK(TextProtocol(), u'foo', u'')
        self.assertRoundTripOK(TextProtocol(), u'caf\xe9', u'\Xe90\c1a')

    def test_no_tabs(self):
        self.assertEqual(TextProtocol().write(u'foo', None), b'foo')
        self.assertEqual(TextProtocol().write(None, u'foo'), b'foo')
        self.assertEqual(TextProtocol().read(b'foo'), (u'foo', None))

        self.assertEqual(TextProtocol().write(u'', None), b'')
        self.assertEqual(TextProtocol().write(None, None), b'')
        self.assertEqual(TextProtocol().write(None, u''), b'')
        self.assertEqual(TextProtocol().read(b''), (u'', None))

    def test_extra_tabs(self):
        self.assertEqual(TextProtocol().write(u'foo', u'bar\tbaz'),
                         b'foo\tbar\tbaz')
        self.assertEqual(TextProtocol().write(u'foo\tbar', u'baz'),
                         b'foo\tbar\tbaz')
        self.assertEqual(TextProtocol().read(b'foo\tbar\tbaz'),
                         (u'foo', u'bar\tbaz'))

    def test_no_strip(self):
        self.assertEqual(TextProtocol().read(b'foo\t \n\n'),
                         (u'foo', u' \n\n'))

    def test_utf_8(self):
        self.assertEqual(TextProtocol().read(b'caf\xc3\xa9\tol\xc3\xa9'),
                         (u'caf\xe9', u'ol\xe9'))

    def test_latin_1_fallback(self):
        # we fall back to latin-1 for the whole line, not individual fields
        self.assertEqual(TextProtocol().read(b'caf\xe9\tol\xc3\xa9'),
                         (u'caf\xe9', u'ol\xc3\xa9'))


class ReprProtocolTestCase(ProtocolTestCase):

    def test_round_trip(self):
        for k, v in REPR_KEYS_AND_VALUES:
            self.assertRoundTripOK(ReprProtocol(), k, v)

    def test_round_trip_with_trailing_tab(self):
        for k, v in REPR_KEYS_AND_VALUES:
            self.assertRoundTripWithTrailingTabOK(ReprProtocol(), k, v)

    def test_uses_repr_format(self):
        KEY = ['a', 1]
        VALUE = {'foo': {'bar': 3}, 'baz': None}
        ENCODED = ('%r\t%r' % (KEY, VALUE)).encode('ascii')

        self.assertEqual((KEY, VALUE), ReprProtocol().read(ENCODED))
        self.assertEqual(ENCODED, ReprProtocol().write(KEY, VALUE))

    def test_bad_data(self):
        self.assertCantDecode(ReprProtocol(), b'{@#$@#!^&*$%^')

    def test_can_encode_point_but_not_decode(self):
        points_encoded = ReprProtocol().write(Point(2, 3), Point(1, 4))
        self.assertCantDecode(ReprProtocol(), points_encoded)


class ReprValueProtocolTestCase(ProtocolTestCase):

    def test_round_trip(self):
        for _, v in REPR_KEYS_AND_VALUES:
            self.assertRoundTripOK(ReprValueProtocol(), None, v)

    def test_round_trip_with_trailing_tab(self):
        for _, v in REPR_KEYS_AND_VALUES:
            self.assertRoundTripWithTrailingTabOK(ReprValueProtocol(), None, v)

    def test_uses_repr_format(self):
        VALUE = {'foo': {'bar': 3}, 'baz': None, 'quz': ['a', 1]}

        self.assertEqual((None, VALUE), ReprValueProtocol().read(repr(VALUE)))
        self.assertEqual(repr(VALUE).encode('ascii'),
                         ReprValueProtocol().write(None, VALUE))

    def test_bad_data(self):
        self.assertCantDecode(ReprValueProtocol(), b'{@#$@#!^&*$%^')

    def test_can_encode_point_but_not_decode(self):
        points_encoded = ReprValueProtocol().write(None, Point(1, 4))
        self.assertCantDecode(ReprValueProtocol(), points_encoded)
